@@ -155,6 +155,20 @@ def tokenize_and_stem(text):
     stems = [stemmer.stem(t) for t in filtered_tokens]
     return stems
 ################################################################################
+def return_stop_words():
+    add_words = ["s"]
+    from sklearn.feature_extraction import text
+    stop_words = text.ENGLISH_STOP_WORDS.union(add_words)
+    from nltk.corpus import stopwords
+    stopWords = set(stopwords.words('english')).union(stop_words)
+    excl =['will',"i'll",'shall',"you'll",'may',"don't","hadn't","hasn't","haven't",
+           "don't","isn't",'if',"mightn't","mustn'","mightn't",'mightn',"needn't",
+           'needn',"needn't",'no','not','shan',"shan't",'shouldn',"shouldn't","wasn't",
+          'wasn','weren',"weren't",'won',"won't",'wouldn',"wouldn't","you'd",'you',
+          "you'd","you'll","you're",'yourself','yourselves']
+    stopWords = [x for x in stopWords if x not in excl]
+    return stopWords    
+################################################################################
 from sklearn.feature_extraction import text 
 from nltk.stem.snowball import SnowballStemmer
 def select_best_nlp_vectorizer(model, data, col, target, metric,
@@ -170,30 +184,17 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     ################################################################################
     """
     nltk.download("popular")
-    add_words = ["s"]
-    from sklearn.feature_extraction import text
-    #stopWords = text.ENGLISH_STOP_WORDS.union(add_words)
-    from nltk.corpus import stopwords
-    stopWords = set(stopwords.words('english')).union(add_words)
-    excl =['will',"i'll",'shall',"you'll",'may',"don't","hadn't","hasn't","haven't",
-           "don't","isn't",'if',"mightn't","mustn'","mightn't",'mightn',"needn't",
-           'needn',"needn't",'no','not','shan',"shan't",'shouldn',"shouldn't","wasn't",
-          'wasn','weren',"weren't",'won',"won't",'wouldn',"wouldn't","you'd",'you',
-          "you'd","you'll","you're",'yourself','yourselves']
-    stopWords = [x for x in stopWords if x not in excl]
+    stopWords = return_stop_words()
     ######################################################
     #### This calculates based on the average number of words in an NLP column how many max_features
-    max_features = int(data[col].map(len).mean()*2)
+    max_df = 0.95
     ######################################################
     if len(data) >= 1000000:
         max_features = 1000
     elif len(data) >= 100000:
-        max_features = 100
+        max_features = 500
     else:
-        try:
-            max_features = data.shape[1]*10
-        except:
-            max_features =  data.shape[0]
+        max_features = int(data[col].map(len).mean()*2)
     print('    A U T O - N L P   P R O C E S S I N G  O N   N L P   C O L U M N = %s ' %col)
     print('#################################################################################')
     print('Generating new features for NLP column = %s using NLP Transformers' %col)
@@ -212,7 +213,7 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     ################################################################
     if modeltype is None or modeltype == '':
         print('Since modeltype is None, Using TFIDF vectorizer with min_df and max_features')
-        tvec = TfidfVectorizer(ngram_range=(1,3), stop_words=stopWords, max_features=int(max_features*0.5), min_df=min_df)
+        tvec = TfidfVectorizer(ngram_range=(1,3), stop_words=stopWords, max_features=max_features, min_df=min_df,max_df=max_df)
         data_dtm =  data[col]
         data_dtm = tvec.fit_transform(data_dtm)
         print('Features: ', data_dtm.shape[1])
@@ -230,8 +231,8 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     print('\n#### First choosing the default Count Vectorizer with 1-3 ngrams and limited features')
     for min_df in sorted(np.linspace(0.10,0.01,10)):
         try:
-            vect_5000 = CountVectorizer(ngram_range=(1, 3), max_features=int(max_features*0.5),
-                                min_df=min_df, binary=False, stop_words=stopWords)
+            vect_5000 = CountVectorizer(ngram_range=(1, 3), max_features=max_features,max_df=max_df,
+                                min_df=2, binary=False, stop_words=stopWords, token_pattern=r'\w{1,}')
             all_vecs[vect_5000], all_models[vect_5000] = tokenize_test_by_metric(model, X_train, X_test, y_train,
                             y_test, target, metric,
                               vect_5000, seed, modeltype)
@@ -239,17 +240,17 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
             break
         except:
             continue
-    print('\n#### Using Vectorizer with max_features < 100 and a low 0.001 min_df with n_gram (2-5)')
+    print('\n#### Using Count Vectorizer with limited max_features and a low 0.001 min_df with n_gram (1-5)')
     ##########################################################################
     ##### It's BEST to use small max_features (50) and a low 0.001 min_df with n_gram (2-5).
     ######  There is no need in that case for stopwords or analyzer since the 2-grams take care of it
     #### Once you do above, there is no difference between count_vectorizer and tfidf_vectorizer
     #### Once u do above, increasing max_features from 50 to even 500 doesn't get you a higher score!
     ##########################################################################
-    vect_lemma = CountVectorizer(
-                                   max_features=int(max_features*0.5), 
-                                   ngram_range=(2, 5), 
-                                    min_df=0.001, 
+    vect_lemma = CountVectorizer(max_df=max_df,
+                                   max_features=max_features, 
+                                   ngram_range=(1, 5), token_pattern=r'\w{1,}',
+                                    min_df=0.001, stop_words=stopWords,
                                    binary=True,
                                     )
     try:
@@ -259,25 +260,27 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     except:
         print('Error: Using CountVectorizer')
 
-    print('\n# Using TFIDF vectorizer with min_df and max_features')
+    print('\n# Using TFIDF vectorizer with min_df=2 and very high max_features')
+    ##### This is based on artificially setting 5GB as being max memory limit for the term-matrix
+    max_features_high = int(250000000/X_train.shape[0])
     if modeltype != 'Regression':
-        tvec = TfidfVectorizer( max_features=int(max_features*0.5), 
-                                stop_words=stopWords, ngram_range=(1, 3), min_df=min_df, binary=True)
+        tvec = TfidfVectorizer( max_features=max_features_high,max_df=max_df, token_pattern=r'\w{1,}',
+                                stop_words=stopWords, ngram_range=(1, 3), min_df=2, binary=True)
     else:
-        tvec = TfidfVectorizer( max_features=int(max_features*0.5), 
-                                stop_words=stopWords, ngram_range=(1, 3), min_df=min_df, binary=False)
+        tvec = TfidfVectorizer( max_features=max_features_high, max_df=max_df, token_pattern=r'\w{1,}',
+                                stop_words=stopWords, ngram_range=(1, 3), min_df=2, binary=False)
     all_vecs[tvec], all_models[tvec] = tokenize_test_by_metric(model, X_train, X_test, y_train,
                                       y_test, target, metric,
                                         tvec, seed, modeltype)
-
-    print('\n# Using TFIDF vectorizer with Snowball Stemming and max_features')
+    max_features_limit = int(tvec.fit_transform(data_dtm).shape[1])
+    print('\n# Using TFIDF vectorizer with Snowball Stemming and limited max_features')
     if modeltype != 'Regression':
-        tvec2 = TfidfVectorizer(max_df=0.80, max_features=int(max_features*0.5),
-                                 min_df=min_df, stop_words=stopWords, binary=True,
+        tvec2 = TfidfVectorizer( max_features=max_features, max_df=max_df, token_pattern=r'\w{1,}',
+                                 min_df=2, stop_words=None, binary=True,
                                  use_idf=True, tokenizer=tokenize_and_stem, ngram_range=(1,3))
     else:
-        tvec2 = TfidfVectorizer(max_df=0.80, max_features=int(max_features*0.5),
-                                 min_df=min_df, stop_words=stopWords, binary=False,
+        tvec2 = TfidfVectorizer( max_features=max_features, max_df=max_df, token_pattern=r'\w{1,}',
+                                 min_df=2, stop_words=None, binary=False,
                                  use_idf=True, tokenizer=tokenize_and_stem, ngram_range=(1,3))
     all_vecs[tvec2], all_models[tvec2] = tokenize_test_by_metric(model, X_train, X_test, y_train,
                                       y_test, target, metric,
@@ -294,7 +297,7 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
             return
     print('\nBest NLP technique selected is: \n%s' %best_vec)
     data_dtm = best_vec.transform(data_dtm)
-    return best_vec, all_models[best_vec], data_dtm, min_df
+    return best_vec, all_models[best_vec], data_dtm, max_features_limit
 
 ############################################################################
 from sklearn.metrics import balanced_accuracy_score,mean_absolute_error,mean_squared_error
@@ -362,6 +365,30 @@ def class_info(classes):
         print("%6s: % 7d  =  % 5.1f%%" % (cls, counts[cls], counts[cls]/total*100))
 
 ###################################################################################
+from sklearn.decomposition import TruncatedSVD
+def select_top_features_from_SVD(X, tsvd, is_train=True, top_n=100):
+    """
+    This program returns the top X features from a TFIDF or CountVectorizer on a dataset.
+    You just need to send in the Vectorized data set X and along with a number denoting
+    how many top features you want back. It will automatically assume you want the top 100.
+    You can change the top X features to any number you want. But it must be less than the
+    number of features in X. Otherwise, it will assume you want all.
+    """
+    X = copy.deepcopy(X)
+    start_time = time.time()
+    #### If the shape of the TFIDF array is huge in the thousands of terms,
+    ####   then you select the top 25 terms in 1-gram and 2-gram that make sense.
+    print('Reducing dimensions from %d term-matrix to %d dimensions using TruncatedSVD...' %(X.shape[1],top_n))
+    if is_train:
+        tsvd = TruncatedSVD(n_components=top_n, 
+                   n_iter=10, 
+                   random_state=3)
+        tsvd = tsvd.fit(X)
+    XA = tsvd.transform(X)
+    print('    Reduced dimensional array shape to %s' %(XA.shape,))
+    print('    Time Taken for Truncated SVD = %0.0f seconds' %(time.time()-start_time) )
+    return XA, tsvd
+###########################################################################################
 from sklearn.feature_extraction.text import TfidfVectorizer
 from collections import defaultdict
 def select_top_features_from_vectorizer(X, vectorizer,top_n=25):
@@ -392,8 +419,7 @@ def select_top_features_from_vectorizer(X, vectorizer,top_n=25):
     if XA.shape[1] <= grams_length:
         #### If the TFIDF array is very small, you just select the entire TFIDF array
         best_features_array = copy.deepcopy(XA)
-        ls = sorted(vectorizer.vocabulary_, key=lambda x: x[0], reverse=False)
-        best_df = pd.DataFrame(best_features_array,columns=ls)
+        best_df = pd.DataFrame(best_features_array.todense(),columns=vectorizer.get_feature_names())
     else:
         #### If the shape of the TFIDF array is huge in the thousands of terms,
         ####   then you select the top 25 terms in 1-gram and 2-gram that make sense.
@@ -431,7 +457,7 @@ from xgboost.sklearn import XGBRegressor
 from sklearn.naive_bayes import GaussianNB
 import xgboost as xgb
 def Auto_NLP(nlp_column, train, test, target, score_type,
-                            seed, modeltype,top_num_features=25, verbose=0):
+                            modeltype,top_num_features=50, verbose=0):
     """
     ##################################################################################
     #### Auto_NLP expects both train and test to be data frames with one NLP column 
@@ -445,9 +471,12 @@ def Auto_NLP(nlp_column, train, test, target, score_type,
     #### You can use top_num_features (default = 25) to control how many features to add.
     ##################################################################################
     """
+    seed = 99
     train = copy.deepcopy(train)
     test = copy.deepcopy(test)
     start_time = time.time()
+    num_classes = len(np.unique(train[target].values))
+    top_num_features = num_classes*top_num_features
     min_df=0.1
     max_depth = 8
     subsample =  0.5
@@ -470,6 +499,14 @@ def Auto_NLP(nlp_column, train, test, target, score_type,
     else:
         print('NLP column must be either a string or a list with one column name in data frame')
         return
+    ########################  S U M M A R Y  C O L U M N S  C R E A T I O N ######################
+    #### Now let's do a combined transformation of NLP column
+    train, nlp_summary_cols = create_summary_of_nlp_cols(train, nlp_column,verbose)
+    nlp_result_columns += nlp_summary_cols
+    print('    Added %d summary columns for counts of words and characters in each row' %len(nlp_summary_cols))
+    if not isinstance(test, str):
+        test, nlp_summary_cols = create_summary_of_nlp_cols(test, nlp_column)
+    ##########################################################################################
     if modeltype.endswith('Classification'):
         #print('Class distribution in Train:')
         #class_info(train[target])
@@ -477,7 +514,7 @@ def Auto_NLP(nlp_column, train, test, target, score_type,
             model = MultinomialNB()
         else:
             model = GaussianNB()
-        best_nlp_vect, model, train_dtm, min_df = select_best_nlp_vectorizer(model, train, nlp_column, target,
+        best_nlp_vect, model, train_dtm, max_features_limit = select_best_nlp_vectorizer(model, train, nlp_column, target,
                     score_type, seed, modeltype,min_df)
     elif modeltype == 'Regression':
         if float(xgb.__version__[0])<1:
@@ -488,20 +525,37 @@ def Auto_NLP(nlp_column, train, test, target, score_type,
                                 colsample_bytree=col_sub_sample,reg_alpha=0.5, reg_lambda=0.5,
                                  seed=1,n_jobs=-1,random_state=1)
         ####   This is where you start to Iterate on Finding Important Features ################
-        best_nlp_vect, model, train_dtm, min_df = select_best_nlp_vectorizer(model, train, nlp_column, target,
+        best_nlp_vect, model, train_dtm, max_features_limit = select_best_nlp_vectorizer(model, train, nlp_column, target,
                             score_type, seed, modeltype,min_df)
     else:
         #### Just complete the transform of NLP column and return the transformed data ####
         model = None
-        best_nlp_vect, model, train_dtm, min_df = select_best_nlp_vectorizer(model, train, nlp_column, target,
+        best_nlp_vect, model, train_dtm, max_features_limit = select_best_nlp_vectorizer(model, train, nlp_column, target,
                             score_type, seed, modeltype,min_df)
     #### Now that the Best VECTORIZER has been selected, transform Train and Test and return vectorized dataframes
-    train_best = select_top_features_from_vectorizer(train_dtm, best_nlp_vect, top_num_features)
+    #### Convert the Feature Array from a Sparse Matrix to a Dense Array #########
+    print('Setting Max Features limit to NLP vectorizer as %d' %max_features_limit)
+    best_nlp_vect.max_features = max_features_limit
+    train_all = best_nlp_vect.fit_transform(train[nlp_column])
+    if train_all.shape[1] <= top_num_features:
+        #### If the TFIDF array is very small, you just select the entire TFIDF array
+        train_best = pd.DataFrame(train_all.todense(),columns=best_nlp_vect.get_feature_names())
+    else:
+        #### For Train data, you don't have to send in an SVD. It will automatically select one and train it
+        best_features_array, tsvd = select_top_features_from_SVD(train_all,'',True)
+        ls = ['svd_dim_'+str(x) for x in range(best_features_array.shape[1])]
+        train_best = pd.DataFrame(best_features_array,columns=ls)
+    #train_best = select_top_features_from_vectorizer(train_dtm, best_nlp_vect, )
     if type(test) != str:
         test_index = test.index
         #### If test data is given, then convert it into a Vectorized frame using best vectorizer
-        test_dtm = best_nlp_vect.transform(test[nlp_column])
-        test_best = select_top_features_from_vectorizer(test_dtm, best_nlp_vect, top_num_features)
+        test_all = best_nlp_vect.transform(test[nlp_column])
+        if test_all.shape[1] <= top_num_features:
+            best_df = pd.DataFrame(test_all.todense(),columns=best_nlp_vect.get_feature_names())
+        else:
+            best_features_array, _ = select_top_features_from_SVD(test_all,tsvd,False)
+            test_best = pd.DataFrame(best_features_array,columns=ls)
+        #test_best = select_top_features_from_vectorizer(test_dtm, best_nlp_vect, top_num_features)
     else:
         test_best = ''
     #### best contains the entire data rows with the top X features of a Vectorizer
@@ -527,10 +581,6 @@ def Auto_NLP(nlp_column, train, test, target, score_type,
         nlp_data = train_nlp.append(test_nlp)
     else:
         nlp_data = copy.deepcopy(train_nlp)
-    #### Now let's do a combined transformation of NLP column
-    nlp_data, nlp_summary_cols = create_summary_of_nlp_cols(nlp_data, nlp_column)
-    nlp_result_columns += nlp_summary_cols
-    print('    Added %d columns for counts of words and characters in each row' %len(nlp_summary_cols))
     #### next create parts-of-speech tagging ####
     if len(nlp_data) <= 10000:
         ### VADER is accurate but very SLOOOWWW. Do not do this for Large data sets ##############
@@ -563,7 +613,7 @@ def Auto_NLP(nlp_column, train, test, target, score_type,
     print('Number of new columns created using NLP = %d' %(len(nlp_result_columns)))
     print('         A U T O   N L P  C O M P L E T E D. Time taken %0.1f minutes' %((time.time()-start_time)/60))
     print('####################################################################################')
-    return train_full, test_full, best_nlp_vect
+    return train_full, test_full, best_nlp_vect, max_features_limit
 ##############################################################################################
 def calculate_line_sentiment(text,senti_type='polarity'):
     review = TextBlob(text)
@@ -620,35 +670,63 @@ def add_sentiment(data, nlp_column):
     print('    Created %d new columns using SentinmentIntensityAnalyzer. Time taken = %d seconds' %(len(cols),time.time()-start_time))
     return data, cols
 ######### Create new columns that provide summary stats of NLP string columns
-def create_summary_of_nlp_cols(data, col,verbose=0):
+def create_summary_of_nlp_cols(data, col, verbose=0):
     """
     Create new columns that provide summary stats of NLP string columns
     This gives us insights into the number of characters we want in our NLP column
     in order for the column to be relevant to the target.
     This can also be a Business question. It may help us in building a better predictive model.
     """
-    data[col+"_num_of_words"] = data[col].apply(lambda x : len(str(x).split())).fillna(0)
+    cols = []
+    stop_words = return_stop_words()
+    # word_count
+    data[col+'_word_count'] = data[col].apply(lambda x: len(str(x).split()))
+    cols.append(col+'_word_count')
+    # unique_word_count
+    data[col+'_unique_word_count'] = data[col].apply(lambda x: len(set(str(x).split())))
+    cols.append(col+'_unique_word_count')
+    # stop_word_count
+    data[col+'_stop_word_count'] = data[col].apply(lambda x: len([w for w in str(x).lower().split() if w in stop_words]))
+    cols.append(col+'_stop_word_count')
+    # url_count
+    data[col+'_url_count'] = data[col].apply(lambda x: len([w for w in str(x).lower().split() if 'http' in w or 'https' in w]))
+    cols.append(col+'_stop_word_count')
+    # mean_word_length
+    data[col+'_mean_word_length'] = data[col].apply(lambda x: int(np.mean([len(w) for w in str(x).split()])))
+    cols.append(col+'_mean_word_length')
+    # char_count
+    data[col+'_char_count'] = data[col].apply(lambda x: len(str(x)))
+    cols.append(col+'_char_count')
+    # punctuation_count
+    data[col+'_punctuation_count'] = data[col].apply(lambda x: len([c for c in str(x) if c in string.punctuation]))
+    cols.append(col+'_punctuation_count')
+    # hashtag_count
+    data[col+'_hashtag_count'] = data[col].apply(lambda x: len([c for c in str(x) if c == '#']))
+    cols.append(col+'_hashtag_count')
+    # mention_count
+    data[col+'_mention_count'] = data[col].apply(lambda x: len([c for c in str(x) if c == '@']))
+    cols.append(col+'_mention_count')
     if verbose >= 1:
-        plot_nlp_column(data[col+"_num_of_words"],"words")
-    data[col+"_num_of_chars"] = data[col].apply(lambda x : len(str(x))).fillna(0)
-    if verbose >= 1:
-        plot_nlp_column(data[col+"_num_of_chars"],"characters")
-    return data, [col+"_num_of_words",col+"_num_of_chars"]
+        plot_nlp_column(data[col+'_unique_word_count'],"Word Count")
+        plot_nlp_column(data[col+'_char_count'],"Character Count")
+    return data, cols
 #############################################################################
 def plot_nlp_column(df_col,label_title):
     """
     We want to know the average number of words per row of text.
     So we first plot the distribution of number of words per text row.
     """
+    color_string = 'bryclg'
     cnt_srs = df_col.value_counts()
     plt.figure(figsize=(12,6))
-    sns.barplot(cnt_srs.index, cnt_srs.values, alpha=0.8, color=color[0])
+    sns.barplot(cnt_srs.index, cnt_srs.values, alpha=0.8, color=color_string[0])
     plt.ylabel('Number of Occurrences', fontsize=12)
-    plt.xlabel('Number of %s in %s' %(label_title,df_col.name),fontsize=12)
+    plt.xlabel('Distribution of %s in newly created %s column' %(label_title,df_col.name),fontsize=12)
     plt.xticks(rotation='vertical')
     plt.show();
 
 def plot_histogram_probability(dist_train, dist_test, label_title):
+    pal = 'bryclg'
     plt.figure(figsize=(15, 10))
     plt.hist(dist_train, bins=200, range=[0, 200], color=pal[2], normed=True, label='train')
     plt.hist(dist_test, bins=200, range=[0, 200], color=pal[1], normed=True, alpha=0.5, label='test')
@@ -659,7 +737,9 @@ def plot_histogram_probability(dist_train, dist_test, label_title):
     plt.show();
 ########################################################################
 module_type = 'Running' if  __name__ == "__main__" else 'Imported'
-version_number = '0.0.16'
+version_number = '0.0.20'
 print("""Imported Auto_NLP version: %s.. Call using:
-     train_nlp, test_nlp, best_nlp_transformer = Auto_NLP(nlp_column, train, test, target, score_type, seed, modeltype)""" %version_number)
+     train_nlp, test_nlp, best_nlp_transformer = Auto_NLP(
+                nlp_column, train, test, target, score_type,
+                modeltype,top_num_features=50, verbose=0)""" %version_number)
 ########################################################################
